import argparse
import hashlib
import hmac
import json
import logging
import re
import sys
import traceback
from typing import Dict, Optional, Text, Tuple
from urllib.parse import urljoin

import act.api
import requests
from act.api.libs import cli

from act.workers.libs import worker

BASEURL = "https://transform.shadowserver.org/api2/"

# Type:Platform/Family.Variant!Suffixes
MS_RE = re.compile(r"(.*?):(.*?)\/(?:([^!.]+))?(?:[!.](\w+))?")
HASH_RE = re.compile(r"^([0-9a-f]{32}|[0-9a-f]{40}|[0-9a-f]{64}|[0-9a-f]{128})$")


def parseargs() -> argparse.ArgumentParser:
    """Extract command lines argument"""

    parser = worker.parseargs("ACT Shadowserver Client")
    parser.add_argument("--apikey", metavar="KEY", help="Shadowserver API key")
    parser.add_argument("--secret", help="Shadowserver secret")
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "--hexdigest",
        action="store_true",
        default=False,
        help="Skip autodetection of type and force lookup as hexdigest",
    )
    group.add_argument(
        "--domain",
        action="store_true",
        default=False,
        help="Skip autodetection of type and force lookup as domain",
    )

    return parser


def query_shadowserver(
    endpoint: Text,
    request: Dict,
    secret: Text,
    proxies: Dict = None,
) -> Dict:

    url = urljoin(BASEURL, endpoint)
    request_string = json.dumps(request)
    secret_bytes = bytes(secret, "latin-1")
    request_bytes = bytes(request_string, "latin-1")
    hmac_generator = hmac.new(secret_bytes, request_bytes, hashlib.sha256)
    hmac2 = hmac_generator.hexdigest()
    response = requests.post(
        url=url, json=request, proxies=proxies, headers={"HMAC2": hmac2}
    )

    if response.status_code == requests.codes.ok:
        return dict(
            results=response.json(),
            response_code=response.status_code,
        )
    else:
        return dict(
            error=response.text,
            response_code=response.status_code,
            proxies=proxies,
        )


def name_extraction(
    actapi: act.api.Act, signature: Text, content_id: Text, output_format: Text = "json"
) -> None:
    """Submit if AV vendor is Microsoft"""
    match = MS_RE.match(signature)

    if match:
        name, toolType = match.groups()[2].lower(), match.groups()[0].lower()
        act.api.helpers.handle_fact(
            actapi.fact("classifiedAs")
            .source("content", content_id)
            .destination("tool", name),
            output_format=output_format,
        )

        if toolType:
            act.api.helpers.handle_fact(
                actapi.fact("classifiedAs")
                .source("tool", name)
                .destination("toolType", toolType),
                output_format=output_format,
            )


def handle_hexdigest(
    actapi: act.api.Act, ioc, apikey, secret, proxies, output_format: Text = "json"
) -> None:
    request = {"apikey": apikey}
    request["sample"] = ioc
    response = query_shadowserver(
        endpoint="research/malware-info",
        request=request,
        secret=secret,
        proxies=proxies,
    )
    try:
        results = response["results"]
    except KeyError:
        logging.error(f"Error handling: {ioc}. Error: {response['error']}")
        sys.exit(1)

    for result in results:
        content_id = result["sha256"]
        for my_hash in ["sha1", "sha256", "sha512", "md5"]:
            act.api.helpers.handle_fact(
                actapi.fact("represents")
                .source("hash", result[my_hash])
                .destination("content", content_id),
                output_format=output_format,
            )

        for r in result["anti_virus"]:
            if r["vendor"] == "Microsoft":
                name_extraction(
                    actapi=actapi, signature=r["signature"], content_id=content_id
                )

        if "domain_name" in result:
            for body in result["domain_name"]:
                if "tag" in body:
                    if "tno-dga-tagged" in body["tag"]:
                        act.api.helpers.handle_facts(
                            act.api.fact.fact_chain(
                                actapi.fact("connectsTo")
                                .source("content", content_id)
                                .destination("uri", "*"),
                                actapi.fact("componentOf")
                                .source("fqdn", body["domain"])
                                .destination("uri", "*"),
                            )
                        )


def handle_domain(
    actapi: act.api.Act, ioc, apikey, secret, proxies, output_format: Text = "json"
) -> None:

    request = {"apikey": apikey}
    request["domain"] = ioc
    response = query_shadowserver(
        endpoint="research/malware-domain",
        request=request,
        secret=secret,
        proxies=proxies,
    )

    try:
        results = response["results"][ioc]
    except KeyError:
        logging.error(f"Error handling: {ioc}. Error: {response['error']}")
        sys.exit(1)

    if not "family" in results:
        logging.warning(f"No family in results: {results}")
        return

    if isinstance(results["family"], str):
        results["family"] = [results["family"]]

    for tool_name in results["family"]:
        for sample in results["samples"]:

            content_id = sample["sha256"]

            act.api.helpers.handle_fact(
                actapi.fact("classifiedAs")
                .source("content", content_id)
                .destination("tool", tool_name),
                output_format=output_format,
            )

            for my_hash in ["sha1", "sha256", "md5"]:
                act.api.helpers.handle_fact(
                    actapi.fact("represents")
                    .source("hash", sample[my_hash])
                    .destination("content", content_id),
                    output_format=output_format,
                )

            for result in sample["anti_virus"]:
                if result["vendor"] == "Microsoft":
                    name_extraction(actapi, result["signature"], content_id)


def run():
    args = cli.handle_args(parseargs())
    actapi = worker.init_act(args)

    if not args.apikey or not args.secret:
        cli.fatal(
            "You must specify --apikey and --secret on command line or in config file"
        )

    proxies = (
        {"http": args.proxy_string, "https": args.proxy_string}
        if args.proxy_string
        else None
    )

    for in_data in sys.stdin:
        in_data = in_data.strip()

        if args.hexdigest or HASH_RE.search(in_data):
            handle_hexdigest(
                actapi,
                in_data,
                args.apikey,
                args.secret,
                proxies,
                output_format=args.output_format,
            )

        elif args.domain:
            handle_domain(
                actapi,
                in_data,
                args.apikey,
                args.secret,
                proxies,
                output_format=args.output_format,
            )


def main_log_error() -> None:
    try:
        run()
    except Exception:
        logging.error(f"Unhandled exception: {traceback.format_exc()}")
        raise


if __name__ == "__main__":
    main_log_error()



""" Tests for worker """
import sys

import _pytest
import act.api
import pytest
from act.api.base import ValidationError
from act.api.helpers import handle_fact
from act.api.libs import cli
from act.types.format import object_format
from act.types.types import object_validates

from act.workers.libs import worker


def test_args_origin_name(monkeypatch: _pytest.monkeypatch.MonkeyPatch) -> None:
    """test argument origin-name"""

    origin_name = "test-origin"

    monkeypatch.setattr(sys, "argv", ["./test-worker.py", "--origin-name", origin_name])

    args = cli.handle_args(worker.parseargs("Test worker"))
    actapi = worker.init_act(args)

    assert actapi.config.origin_name == origin_name

    fact = (
        actapi.fact("mentions").source("report", "xyz").destination("fqdn", "test.com")
    )

    assert fact.origin.name == origin_name


def test_args_origin_id(monkeypatch: _pytest.monkeypatch.MonkeyPatch) -> None:
    """test argument origin-id"""

    origin_id = "00000000-0000-0000-0000-000000000001"

    monkeypatch.setattr(sys, "argv", ["./test-worker.py", "--origin-id", origin_id])

    args = cli.handle_args(worker.parseargs("Test worker"))
    actapi = worker.init_act(args)

    assert actapi.config.origin_id == origin_id

    fact = (
        actapi.fact("mentions").source("report", "xyz").destination("fqdn", "test.com")
    )

    assert fact.origin.id == origin_id


def test_validator_no_strict(caplog) -> None:
    api = act.api.Act(
        "",
        None,
        object_formatter=object_format,
        object_validator=object_validates,
    )

    # Should return None if fact does not validate
    fact = handle_fact(
        api.fact("mentions")
        .source("report", "xyz")
        .destination("uri", "X7f://cve-2014-0224")
    )

    assert fact is None

    assert "Destination object does not validate:" in caplog.text


def test_validate_same_object() -> None:
    api = act.api.Act(
        "",
        None,
        strict_validator=True,
        object_formatter=object_format,
        object_validator=object_validates,
    )

    act.api.helpers.handle_fact.cache_clear()

    with pytest.raises(ValidationError, match=r"Source object can not be equal to.*"):
        handle_fact(
            api.fact("mentions").source("report", "xyz").destination("report", "xyz")
        )


def test_validator_no_validator(caplog) -> None:
    api = act.api.Act("", None)
    act.api.helpers.handle_fact.cache_clear()

    # Should return None if fact does not validate
    fact = handle_fact(
        api.fact("mentions")
        .source("report", "xyz")
        .destination("uri", "X7f://cve-2014-0224")
    )

    # No validator is specified so the above should return a fact
    assert fact is not None

    # Should not log errors
    assert caplog.text == ""


def test_validator_strict() -> None:
    api = act.api.Act(
        "",
        None,
        strict_validator=True,
        object_formatter=object_format,
        object_validator=object_validates,
    )
    act.api.helpers.handle_fact.cache_clear()

    with pytest.raises(
        ValidationError, match=r"Destination object does not validate.*"
    ):

        handle_fact(
            api.fact("mentions")
            .source("report", "xyz")
            .destination("uri", ".X7f://cve-2014-0224")
        )


def test_format() -> None:
    api = act.api.Act(
        "",
        None,
        object_formatter=object_format,
        object_validator=object_validates,
    )
    act.api.helpers.handle_fact.cache_clear()

    ta_alias = handle_fact(
        api.fact("alias")
        .source("threatActor", "APT29")
        .destination("threatActor", "Cozy Bear")
    )

    assert ta_alias.source_object.value == "apt29"
    assert ta_alias.destination_object.value == "cozy bear"



#!/usr/bin/env python3

"""iSight worker for the ACT platform

Copyright 2018 the ACT project <opensource@mnemonic.no>

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted, provided that the above
copyright notice and this permission notice appear in all copies.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.

requirements:

"""


import argparse
import contextlib
import datetime
import email
import hashlib
import hmac
import json
import logging
import os
import socket
import time
import traceback
import warnings
from functools import partialmethod
from logging import error
from typing import Dict, Generator, Text

import act.api
import requests
from act.api.helpers import handle_fact, handle_facts, handle_uri
from act.api.libs import cli

from act.workers.libs import worker


def parseargs() -> argparse.ArgumentParser:
    """Extract command lines argument"""

    parser = worker.parseargs("ACT iSight Client")
    parser.add_argument("--privatekey", metavar="PRIVATEKEY", help="iSight API key")
    parser.add_argument("--publickey", metavar="PUBLICKEY", help="iSight API key")
    parser.add_argument("--debugdir", metavar="DIR", help="Dump directory for output")
    parser.add_argument(
        "--days", default="1", help="How many days back to look for data"
    )
    parser.add_argument(
        "--root", default="https://api.isightpartners.com", help="api endpoint"
    )

    return parser


def is_valid_ipv4_address(address: Text) -> bool:
    try:
        socket.inet_pton(socket.AF_INET, address)
    except socket.error:  # not a valid address
        return False
    return True


def is_valid_ipv6_address(address: Text) -> bool:
    try:
        socket.inet_pton(socket.AF_INET6, address)
    except socket.error:  # not a valid address
        return False
    return True


OBJECT_MAP = {
    "sourceDomain": lambda x: "fqdn",
    "sourceIP": lambda x: "ipv4"
    if is_valid_ipv4_address(x)
    else "ipv6"
    if is_valid_ipv6_address(x)
    else None,
    "fuzzyHash": lambda x: "hash",
    "md5": lambda x: "hash",
    "sha1": lambda x: "hash",
    "sha256": lambda x: "hash",
    "asn": lambda x: "asn",
    "cidr": lambda x: "ipv4Network",
    "domain": lambda x: "fqdn",
    "ip": lambda x: "ipv4"
    if is_valid_ipv4_address(x)
    else "ipv6"
    if is_valid_ipv6_address(x)
    else None,
    "malwareFamily": lambda x: "tool",
    "actor": lambda x: "threatActor",
}


def main() -> None:
    """main function"""

    # Look for default ini file in "/etc/actworkers.ini" and ~/config/actworkers/actworkers.ini
    # (or replace .config with $XDG_CONFIG_DIR if set)
    args = cli.handle_args(parseargs())

    actapi = worker.init_act(args)

    if not (args.privatekey and args.publickey):
        cli.fatal(
            "You must specify --privatekey and --publickey on command line or in config file"
        )

    proxies = (
        {"http": args.proxy_string, "https": args.proxy_string}
        if args.proxy_string
        else None
    )

    iSightHandler = ISightAPIRequestHandler(args.root, args.privatekey, args.publickey)
    data = iSightHandler.indicators(days=args.days, proxies=proxies)

    if "success" not in data or not data["success"]:
        logging.error(
            "Unable to download from isight API [%s]",
            data["message"] if "message" in data else "NA",
        )
        return

    timestamp = int(time.time())
    ### DEBUG -- dump json to disc for each run
    if args.debugdir:
        with open(
            os.path.join(args.debugdir, "error-{0}.json".format(timestamp)), "w"
        ) as f:
            json.dump(data, f)

    for i, dp in enumerate(data["message"]):
        ### --- Handle mentions facts
        # Create report ID from the url (same approach as for feeds) and title to this ID.
        reportID = hashlib.sha256(dp["webLink"].encode("utf8")).hexdigest()
        handle_fact(actapi.fact("name", dp["title"]).source("report", reportID))
        for obj in OBJECT_MAP:  # run through all fields that we want to mention
            if obj in dp and dp[obj]:  # if the report contains data in the field
                factType = OBJECT_MAP[obj](dp[obj])  # translate to ACT fact type
                handle_fact(
                    actapi.fact("mentions")  # and create fact from field
                    .source("report", reportID)
                    .destination(factType, dp[obj].lower())
                )
        if dp["url"]:
            handle_fact(
                actapi.fact("mentions")
                .source("report", reportID)
                .destination("uri", dp["url"])
            )
            handle_uri(actapi, dp["url"])
        ### --- IP -> malwareFamily
        if dp["malwareFamily"] and dp["ip"]:
            handle_facts(
                act.api.fact.fact_chain(
                    actapi.fact("connectsTo")
                    .source("content", "*")
                    .destination("uri", "*"),
                    actapi.fact("componentOf")
                    .source("ipv4", dp["ip"])
                    .destination("uri", "*"),
                    actapi.fact("classifiedAs")
                    .source("content", "*")
                    .destination("tool", dp["malwareFamily"]),
                )
            )
        ### --- URL -> malwareFamily
        elif dp["networkType"] == "url" and dp["malwareFamily"]:
            handle_uri(actapi, dp["url"])
            handle_facts(
                act.api.fact.fact_chain(
                    actapi.fact("connectsTo")
                    .source("content", "*")
                    .destination("uri", dp["url"]),
                    actapi.fact("classifiedAs")
                    .source("content", "*")
                    .destination("tool", dp["malwareFamily"]),
                )
            )
        ### --- FQDN -> malwareFamily
        elif dp["networkType"] == "network" and dp["domain"] and dp["malwareFamily"]:
            handle_facts(
                act.api.fact.fact_chain(
                    actapi.fact("connectsTo")
                    .source("content", "*")
                    .destination("uri", "*"),
                    actapi.fact("componentOf")
                    .source("fqdn", dp["domain"])
                    .destination("uri", "*"),
                    actapi.fact("classifiedAs")
                    .source("content", "*")
                    .destination("tool", dp["malwareFamily"]),
                )
            )
        ### --- hash -> malwareFamily
        elif (
            dp["fileType"]
            and dp["malwareFamily"]
            and (dp["sha1"] or dp["sha256"] or dp["md5"])
        ):
            for digest_type in ["md5", "sha1", "sha256"]:
                ### In some cases the iSight api does not return a sha256 hashdigest
                ### so we need to make a chain through a placeholder content
                if not dp["sha256"]:
                    if dp[digest_type]:
                        handle_facts(
                            act.api.fact.fact_chain(
                                actapi.fact("represents")
                                .source("hash", dp[digest_type])
                                .destination("content", "*"),
                                actapi.fact("classifiedAs")
                                .source("content", "*")
                                .destination("tool", dp["malwareFamily"]),
                            )
                        )
                else:  ## There is a sha256, so we do _not_ need a chain
                    if dp[digest_type]:
                        handle_fact(
                            actapi.fact("classifiedAs")
                            .source("content", dp["sha256"])
                            .destination("tool", dp["malwareFamily"])
                        )

                        handle_fact(
                            actapi.fact("represents")
                            .source("hash", dp[digest_type])
                            .destination("content", dp["sha256"])
                        )

        ### -- Hash --> actor
        elif (
            dp["fileType"] and dp["actor"] and (dp["sha1"] or dp["sha256"] or dp["md5"])
        ):
            for digest_type in ["md5", "sha1", "sha256"]:
                ### In some cases the iSight api does not return a sha256 hashdigest
                ### so we need to make a chain through a placeholder content
                if not dp["sha256"]:
                    if dp[digest_type]:
                        handle_facts(
                            act.api.fact.fact_chain(
                                actapi.fact("represents")
                                .source("hash", dp[digest_type])
                                .destination("content", "*"),
                                actapi.fact("observedIn")
                                .source("content", "*")
                                .destination("incident", "*"),
                                actapi.fact("attributedTo")
                                .source("incident", "*")
                                .destination("threatActor", dp["actor"]),
                            )
                        )
                else:  ## There is a sha256, so we do _not_ need a chain between all the way from hexdigest
                    if dp[digest_type]:
                        handle_fact(
                            actapi.fact("represents")
                            .source("hash", dp[digest_type])
                            .destination("content", dp["sha256"])
                        )

                        handle_facts(
                            act.api.fact.fact_chain(
                                actapi.fact("observedIn")
                                .source("content", dp["sha256"])
                                .destination("incident", "*"),
                                actapi.fact("attributedTo")
                                .source("incident", "*")
                                .destination("threatActor", dp["actor"]),
                            )
                        )
        ### We do have a sha256 of a file (but possibly nothing else). Add the content to hexdigest facts
        elif dp["fileType"] and dp["sha256"]:
            for digest in ["sha1", "md5", "sha256"]:
                if dp[digest]:
                    handle_fact(
                        actapi.fact("represents")
                        .source("hash", dp[digest])
                        .destination("content", dp["sha256"])
                    )
            if args.debugdir:
                fields = [
                    k
                    for k, v in dp.items()
                    if v
                    and k
                    not in [
                        "reportId",
                        "title",
                        "ThreatScape",
                        "audience",
                        "intelligenceType",
                        "publishDate",
                        "reportLink",
                        "webLink",
                    ]
                ]
                logging.error(
                    "[%s] Extra fields while handeling index[%s] '%s'",
                    timestamp,
                    i,
                    ", ".join(fields),
                )

        ### -- DEBUG!
        else:
            if args.debugdir:
                fields = [
                    k
                    for k, v in dp.items()
                    if v
                    and k
                    not in [
                        "reportId",
                        "title",
                        "ThreatScape",
                        "audience",
                        "intelligenceType",
                        "publishDate",
                        "reportLink",
                        "webLink",
                    ]
                ]
                logging.error(
                    "[%s] Unable to handle index[%s] with fields '%s'",
                    timestamp,
                    i,
                    ", ".join(fields),
                )


## -----------------------------------------


class ISightAPIRequestHandler(object):

    INDICATORS = "/view/indicators"

    def __init__(self, root: Text, private_key: Text, public_key: Text) -> None:
        """Create a new iSight api handler with api root and keys"""

        self.URL = root
        self.public_key = public_key
        self.private_key = private_key
        self.accept_version = "2.6"

    def indicators(self, days: int = 1, proxies: Dict = None) -> Dict:
        """Download indicators last X days"""

        toTS = int(datetime.datetime.now().timestamp())
        fromTS = int(
            (datetime.datetime.now() - datetime.timedelta(days=int(days))).timestamp()
        )

        time_stamp = email.utils.formatdate(localtime=True)
        ENDPOINT = self.INDICATORS + "?startDate={0}&endDate={1}".format(fromTS, toTS)
        accept_header = "application/json"
        new_data = ENDPOINT + self.accept_version + accept_header + time_stamp

        key = bytearray()
        key.extend(map(ord, self.private_key))
        hashed = hmac.new(key, new_data.encode("utf-8"), hashlib.sha256)

        headers = {
            "Accept": accept_header,
            "Accept-Version": self.accept_version,
            "X-Auth": self.public_key,
            "X-Auth-Hash": hashed.hexdigest(),
            "Date": time_stamp,
        }

        r = requests.get(self.URL + ENDPOINT, headers=headers, proxies=proxies)
        status_code = r.status_code

        if status_code == 200:
            return json.loads(r.text)
        else:
            logging.error(r.text)
            return {"message": r.text}


@contextlib.contextmanager
def no_ssl_verification() -> Generator[None, None, None]:
    """Monkey patch request to default to no verification of ssl"""

    old_request = requests.Session.request
    requests.Session.request = partialmethod(old_request, verify=False)  # type: ignore

    warnings.filterwarnings("ignore", "Unverified HTTPS request")
    yield
    warnings.resetwarnings()

    requests.Session.request = old_request  # type: ignore


def main_log_error() -> None:
    try:
        main()
    except Exception:
        error("Unhandled exception: {}".format(traceback.format_exc()))
        raise


if __name__ == "__main__":
    main_log_error()



#!/usr/bin/env python3

"""General ACT backend uploader. Reads facts as JSON
from the stdin, uploading accordingly"""

import argparse
import json
import sys
import time
import traceback
from logging import error, warning
from typing import Iterable, Text

import act.api
from act.api.libs import cli

from act.workers.libs import worker


def parseargs() -> argparse.ArgumentParser:
    """Parse arguments"""
    parser = worker.parseargs("Generic uploader")
    parser.add_argument(
        "--allow-default-origin", action="store_true", help="Allow facts without origin"
    )
    parser.add_argument(
        "--timing", action="store_true", help="Add timing operations at warn level"
    )

    parser.add_argument(
        "--no-exit-on-error", action="store_true", help="Log errors and continue"
    )

    parser.add_argument(
        "--no-format",
        action="store_true",
        help="Do not format facts with default formatter",
    )

    parser.add_argument(
        "--no-validate",
        action="store_true",
        help="Do not validate facts with default types",
    )

    return parser


def uploader(
    actapi: act.api.Act,
    iterator: Iterable[Text],
    timing: bool = False,
    allow_default_origin: bool = False,
    no_exit_on_error: bool = False,
) -> None:
    """Process stdin, parse each separat line as a JSON structure and
    register a fact based on the structure. The form of input should
    be the on the form accepted by the ACT Rest API fact API."""

    handle_fact_time = []
    origins = set()
    facts = {}
    metafacts = set()

    for line in iterator:
        try:
            data = json.loads(line)
        except json.decoder.JSONDecodeError:
            error("Error decoding json: %s" % line)
            continue

        # Create either Fact or MetaFact, if we have the inReferenceTo field populated or not
        fact = (
            actapi.meta_fact(**data) if "inReferenceTo" in data else actapi.fact(**data)
        )

        if not allow_default_origin:
            if not (fact.origin and (fact.origin.id or fact.origin.name)):
                error("Origin not specified: %s", fact)
                continue

        started = time.time()

        # If this is a meta fact, save for later upload
        if isinstance(fact, act.api.fact.MetaFact):
            metafacts.add(fact)
            continue

        # Handle fact
        # Fact will be added to the platform if base_url is configured
        # The returned fact will be populated with a fact id
        # and default values from the platform
        try:
            fact_copy = act.api.helpers.handle_fact(fact)
        except act.api.base.ValidationError as err:
            warning("ValidationError while storing objects: %s" % err)
            continue
        except act.api.base.ResponseError as err:
            error("ResponseError while storing objects: %s" % err)

            if not no_exit_on_error:
                sys.exit(1)

        # Add cache for facts to be used to populate id in meta facts
        if isinstance(fact, act.api.fact.Fact):
            facts[fact] = fact_copy.id

        # Keep list of all known origins
        if fact.origin:
            origins.add(fact.origin.name)

        time_spent = time.time() - started
        handle_fact_time.append(time_spent)
        if timing:
            warning("Handle fact time: %s", round(time_spent, 2))

    for fact in metafacts:
        if fact.in_reference_to not in facts:
            error(
                "Recieved metafact that references unknown fact. "
                "This error can occur if the fact is not submitted in the same file "
                "or the fact fails to validate."
            )
            continue

        # Update ID of referenced fact from cache
        fact.in_reference_to.id = facts[fact.in_reference_to]
        try:
            act.api.helpers.handle_fact(fact)
        except act.api.base.ValidationError as err:
            warning("ValidationError while storing objects: %s" % err)
            continue
        except act.api.base.ResponseError as err:
            error("ResponseError while storing objects: %s" % err)
            sys.exit(1)

    if timing and handle_fact_time:
        warning(
            "Total time (count:%s,total:%s,mean:%s,min:%s,max:%s,origins:%s)",
            len(handle_fact_time),
            round(sum(handle_fact_time), 2),
            round(sum(handle_fact_time) / len(handle_fact_time), 2),
            round(min(handle_fact_time), 2),
            round(max(handle_fact_time), 2),
            "+".join(origins),
        )


def main_log_error() -> None:
    "Call main() and log all exceptions as errors"
    try:
        # Look for default ini file in "/etc/actworkers.ini" and ~/config/actworkers/actworkers.ini
        # (or replace .config with $XDG_CONFIG_DIR if set)
        args = cli.handle_args(parseargs())
        actapi = worker.init_act(
            args, no_format=args.no_format, no_validate=args.no_validate
        )

        uploader(
            actapi,
            sys.stdin,
            args.timing,
            args.allow_default_origin,
            args.no_exit_on_error,
        )
    except Exception:
        error("Unhandled exception: {}".format(traceback.format_exc()))
        raise


if __name__ == "__main__":
    main_log_error()



#!/usr/bin/env python3

"""Veris worker for the ACT platform

Copyright 2019 the ACT project <opensource@mnemonic.no>

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted, provided that the above
copyright notice and this permission notice appear in all copies.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
"""

import argparse
import csv
import io
import json
import os
import re
import sys
import traceback
import zipfile
from logging import error, warning
from typing import Any, Dict, Optional, Text, Union, cast

import requests

import act.api
from act.api.helpers import handle_fact, handle_facts
from act.api.libs import cli
from act.workers.libs import urlcache, worker

VERSION = "0.1"
ISO_3166_FILE = (
    "https://raw.githubusercontent.com/lukes/"
    + "ISO-3166-Countries-with-Regional-Codes/master/all/all.json"
)


def get_cn_map(filename: Text) -> Dict:
    """
    Read file with county information (ISO 3166 from filename)
    return map with country code (e.g. "NO") as key, and Country
    Name (e.g. "Norway" as value)
    """
    cn_map = {}

    for c_map in json.loads(open(filename, "r", encoding="utf8").read()):
        cn_map[c_map["alpha-2"]] = c_map["name"]

    return cn_map


def parseargs() -> argparse.ArgumentParser:
    """Parse arguments"""
    parser = worker.parseargs("Shadowserver ASN enrichment")
    parser.add_argument(
        "--country-codes",
        help="Should point to file downloaded from {}".format(ISO_3166_FILE),
    )
    parser.add_argument(
        "--veris-campaign", help="Read mapping of veris campaign from (CSV) file"
    )
    parser.add_argument(
        "--threat-actor-variety",
        help="Varieties to use as Threat Actors",
        default="Activist, Organized crime, Nation-state",
    )
    parser.add_argument(
        "--hash-url-matching",
        help="Download and hash references matching regular expression",
        default=r"^(.*pdf)$",
    )
    parser.add_argument(
        "--veris-prefix",
        help='Prefix for incidents and campaign IDs. E.g use "VCDB" for Veris Community Database"',
    )
    parser.add_argument("--veris-url", help="Read veris incidents from URL.")
    parser.add_argument("--veris-file", help="Read veris incidents from File.")
    parser.add_argument(
        "--stdin", action="store_true", help="Read veris incidents on stdin."
    )
    return parser


def handle_reports(
    config: Dict[Text, Any], incident: Dict[Text, Any], incident_id: Text
) -> None:
    """
    Extract all references in incident. For each (URL-)reference, check whether we should
    download the content and creaa a sha256 hash digest of it (must match config["hash_url_matching"]).
    The sha256 hash is also cached locally.
    """

    # Split references by ";"
    references = [
        ref.strip() for ref in incident.get("reference", "").split(";") if ref.strip()
    ]

    if references:
        for ref in references:
            if not re.search(config["hash_url_matching"], ref):
                continue

            report_hash = config["db_cache"].query_download_update(ref)

            if report_hash:
                handle_fact(
                    config["actapi"]
                    .fact("mentions")
                    .source("report", report_hash)
                    .destination("incident", incident_id),
                    output_format=config["output_format"],
                )


def handle_organizations(
    config: Dict[Text, Any], incident: Dict[Text, Any], incident_id: Text
) -> None:
    """
    Create facts:
        * incident -targets-> organization
        * organization -locatedIn-> country
    """
    victim = incident.get("victim", {}).get("victim_id")

    if not victim:
        return

    handle_fact(
        config["actapi"]
        .fact("targets")
        .source("incident", incident_id)
        .destination("organization", victim),
        output_format=config["output_format"],
    )

    for country_code in incident.get("victim", {}).get("country", []):
        country = config["cn_map"].get(country_code)

        if country:
            handle_fact(
                config["actapi"]
                .fact("locatedIn")
                .source("organization", victim)
                .destination("country", country),
                output_format=config["output_format"],
            )


def handle_threat_actor(
    config: Dict[Text, Any], incident: Dict[Text, Any], incident_id: Text
) -> None:
    """
    Creat facts from actor->external, where the variety includes at least on variety specified in config["threat_actor_variety"]

    """

    external_actor = incident.get("actor", {}).get("external", {})

    if not external_actor:
        return  # No threat actors

    # Varieties are "tags" on actors and we do not want to include all type of actors
    # Make sure at least one of the varieties are in the list of the configured varieties to include
    if any(
        [
            variety in config["threat_actor_variety"]
            for variety in external_actor.get("variety", [])
        ]
    ):
        threat_actors = [
            ta.strip()
            for ta in incident.get("actor", {}).get("external", {}).get("name", [])
        ]

        for ta in threat_actors:
            handle_fact(
                config["actapi"]
                .fact("attributedTo")
                .source("incident", incident_id)
                .destination("threatActor", ta),
                output_format=config["output_format"],
            )


def handle_tool(
    config: Dict[Text, Any], incident: Dict[Text, Any], incident_id: Text
) -> None:
    "Create content -classifiedAs-> tool, and fact chain from content to incident."

    # Both "," and ";" are used to separate tools :(

    tools = [
        malware.strip().lower()
        for malware in re.split(
            r";|,", incident.get("action", {}).get("malware", {}).get("name", "")
        )
        if malware
    ]

    for tool in tools:
        handle_facts(
            act.api.fact.fact_chain(
                config["actapi"]
                .fact("classifiedAs")
                .source("content", "*")
                .destination("tool", tool),
                config["actapi"]
                .fact("observedIn")
                .source("content", "*")
                .destination("incident", incident_id),
            ), 
            output_format=config["output_format"]
        )


def handle_campaign(
    config: Dict[Text, Any], incident: Dict[Text, Any], incident_id: Text
) -> None:
    """
    Create incident -attributedTo-> campaign facts
    If we have a mapping from campaign (UUID) to name, a campaign -name-> will also be created
    """

    campaign = (
        "{}-{}".format(config["veris_prefix"], incident["campaign_id"])
        if incident.get("campaign_id")
        else None
    )

    if not campaign:
        return

    handle_fact(
        config["actapi"]
        .fact("attributedTo")
        .source("incident", incident_id)
        .destination("campaign", campaign),
        output_format=config["output_format"],
    )

    name = config["campaign_map"].get(format(campaign))

    if name:
        handle_fact(
            config["actapi"].fact("name", name).source("campaign", campaign),
            output_format=config["output_format"],
        )

    else:
        warning(
            "No name found for campaign {}. Make sure veris-campaign is provided and the ID is included in csv field (without prefix)".format(
                campaign
            )
        )


def handle_incident(config: Dict[Text, Any], incident: Dict[Text, Any]) -> None:
    """handle veris incidents"""

    incident_id = "{}-{}".format(config["veris_prefix"], incident["incident_id"])

    handle_reports(config, incident, incident_id)
    handle_organizations(config, incident, incident_id)
    handle_tool(config, incident, incident_id)
    handle_threat_actor(config, incident, incident_id)
    handle_campaign(config, incident, incident_id)


def handle_zip_file(config: Dict[Text, Any], zfile: Union[Text, io.BytesIO]) -> None:
    "Read incident form all (json) files in zip file"

    # ZipFile accepts both file names (str) and file like objects
    zf = zipfile.ZipFile(zfile, "r")

    for fileinfo in zf.infolist():
        for incident in json.loads(zf.read(fileinfo).decode("utf-8")):
            handle_incident(config, incident)


def process(config: Dict[Text, Any]) -> None:
    "Process inicdent from stdin, url or file. URL and File supports zipped files in zip-format."

    # Read incident from stdin
    if config["stdin"]:
        for incident in sys.stdin.read().split("\n"):
            handle_incident(config, json.loads(incident))

        return

    # Download incidents from URL
    if config["veris_url"]:
        req = requests.get(
            config["veris_url"],
            proxies=config["proxies"],
            timeout=config["http_timeout"],
        )
        if config["veris_url"].endswith(".zip"):
            handle_zip_file(config, io.BytesIO(req.content))
        else:
            for incident in req.json():
                handle_incident(config, cast(Dict, incident))
        return

    # Read incidents from file
    if config["veris_file"]:
        if config["veris_file"].endswith(".zip"):
            handle_zip_file(config, config["veris_file"])
        else:
            with open(config["veris_file"]) as f:
                for incident in json.loads(f.read()):
                    handle_incident(config, cast(Dict, incident))
        return

    error("Must specifiy either --stdin, --veris-url or --veris-file")


def get_campaigns(veris_prefix: Text, filename: Text) -> Optional[Dict[Text, Text]]:
    """
    Get mapping from campaign (UUID) to name
    """
    with open(filename) as csvfile:
        return {
            "{}-{}".format(veris_prefix, row[0]): row[1]
            for row in csv.reader(csvfile, delimiter=",")
        }


def main() -> None:
    """main function"""

    # Look for default ini file in "/etc/actworkers.ini" and ~/config/actworkers/actworkers.ini
    # (or replace .config with $XDG_CONFIG_DIR if set)
    args = cli.handle_args(parseargs())
    actapi = worker.init_act(args)

    if not args.country_codes:
        cli.fatal("You must specify --country-codes on command line or in config file")

    if not args.veris_prefix:
        cli.fatal("You must specify --veris-prefix")

    if not (args.veris_url or args.veris_file or args.stdin):
        cli.fatal("You must specify --veris-url, --veris-file or --stdin")

    args.veris_prefix = args.veris_prefix.upper()

    if not os.path.isfile(args.country_codes):
        cli.fatal(
            "Country/region file not found at specified location: {}".format(
                args.country_codes
            ),
            2,
        )

    args.threat_actor_variety = [
        variety.strip() for variety in args.threat_actor_variety.split(",")
    ]

    # Configuration object that will be passed around to functions
    config = {
        # act API
        "actapi": actapi,
        # Map of CC -> Country Name
        "cn_map": get_cn_map(args.country_codes),
        # Map of CC -> Country Name
        "campaign_map": get_campaigns(args.veris_prefix, args.veris_campaign)
        if args.veris_campaign
        else {},
        # Cache of url > sha256
        "db_cache": urlcache.URLCache(
            requests_common_kwargs={
                "proxies": {"http": args.proxy_string, "https": args.proxy_string},
                "timeout": args.http_timeout,
            }
        ),
        "proxies": {"http": args.proxy_string, "https": args.proxy_string}
        if args.proxy_string
        else None,
    }

    # Add all arguments from args to config
    config.update(vars(args))

    process(config)


def main_log_error() -> None:
    "Main function. Log all exceptions to error"
    try:
        main()
    except Exception:
        error("Unhandled exception: {}".format(traceback.format_exc()))
        raise


if __name__ == "__main__":
    main_log_error()



#!/usr/bin/env python3

"""VirusTotal worker for the ACT platform

Copyright 2018 the ACT project <opensource@mnemonic.no>

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted, provided that the above
copyright notice and this permission notice appear in all copies.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.

requirements:

    https://github.com/blacktop/virustotal-api

    pip install virustotal-api
"""

import argparse
import collections
import contextlib
import ipaddress
import logging
import os
import re
import sys
import traceback
import urllib.parse
import warnings
from functools import partialmethod
from logging import error, info, warning
from typing import Generator, List, Optional, Text, Tuple, Set

import requests

import act.api
from act.api.libs import cli
from act.workers.libs import worker
from virus_total_apis import PublicApi as VirusTotalApi

ADWARE_OVERRIDES = ["opencandy", "monetize", "adload", "somoto"]

HASH_RE = re.compile(r"^([0-9a-f]{32}|[0-9a-f]{40}|[0-9a-f]{64}|[0-9a-f]{128})$")

# Type:Platform/Family.Variant!Suffixes
MS_RE = re.compile(r"(.*?):(.*?)\/(?:([^!.]+))?(?:[!.](\w+))?")

# [Prefix:]Behaviour.Platform.Name[.Variant]
KASPERSKY_RE = re.compile(r"((.+?):)?(.+?)\.(.+?)\.([^.]+)(\.(.+))?")

# Extract CVE
CVE_RE = re.compile(r"(CVE-\d+-\d+)")

# <Threat Type>.<Platform>.<Malware Family>.<Variant>.<Other info*>
# *Optional
TREND_RE = re.compile(r"(.+?)\.(.+?)\.(.+?)\.(.+?)(\.(.+))?")


def parseargs() -> argparse.ArgumentParser:
    """Extract command lines argument"""

    parser = worker.parseargs("ACT VT Client")
    parser.add_argument("--apikey", metavar="KEY", help="VirusTotal API key")
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "--hexdigest",
        action="store_true",
        default=False,
        help="Skip autodetection of type and force lookup as hexdigest",
    )
    group.add_argument(
        "--ip",
        action="store_true",
        default=False,
        help="Skip autodetection of type and force lookup as IP address",
    )
    group.add_argument(
        "--domain",
        action="store_true",
        default=False,
        help="Skip autodetection of type and force lookup as domain",
    )

    return parser


def name_extraction(engine: Text, body: dict) -> Optional[Tuple[Text, Optional[Text]]]:
    """Extract the name from certain AV engines based on regular
    expression matching"""

    if engine == "Microsoft":
        match = MS_RE.match(body["result"])
        if match:
            return match.groups()[2].lower(), match.groups()[0].lower()

    if engine == "Kaspersky":
        match = KASPERSKY_RE.match(body["result"])
        if match:
            # Kaspersky does not allways include toolType in the naming scheme
            if match.groups()[1]:
                # Extract field 1 (inner match group) as field 0 (outer match group)
                # contains the ending ':'
                toolType: Optional[Text] = match.groups()[1].lower()
            else:
                toolType = None
            return match.groups()[4].lower(), toolType

    if engine == "TrendMicro":
        match = TREND_RE.match(body["result"])
        if match:
            return match.groups()[2].lower(), match.groups()[0].lower()

    return None


def cve_extraction(body: dict) -> Set[Text]:
    """Extract any CVE names"""

    return {cve.lower() for cve in CVE_RE.findall(body["result"])}


def is_adware(text: Text) -> bool:
    """Test for adware signature using heuristics in ADWARE_OVERRIDES"""

    for adware_override in ADWARE_OVERRIDES:
        if adware_override in text:
            return True
    return False


def handle_hexdigest(
    actapi: act.api.Act,
    vtapi: VirusTotalApi,
    hexdigest: Text,
    cache: dict = {},
    output_format: Text = "json",
) -> None:
    """Read hexdigest from stdin, query VirusTotal and
    output a JSON text readable by generic_uploader.py"""

    if hexdigest in cache:
        return

    cache["hexdigest"] = True

    names: Set[Tuple[Text, Optional[Text]]] = set()
    cves: Set[Text] = set()

    with no_ssl_verification():
        response = vtapi.get_file_report(hexdigest)

    if "results" not in response:
        logging.error("%s in handle_hexdigest for %s", response, hexdigest)
        sys.exit(1)

    if "scans" not in response["results"]:
        # VirusTotal has not seend this hexdigest before
        return

    for engine, body in response["results"]["scans"].items():
        if not body["detected"]:
            continue

        cves.update(cve_extraction(body))

        ext_res = name_extraction(engine, body)
        if ext_res:
            names.add((ext_res[0], ext_res[1]))

        res = body["result"].lower()

        if is_adware(res):
            names.add(("adware", "adware"))

    results = response["results"]
    content_id = results["sha256"]
    for myhash in ["sha1", "sha256", "md5"]:
        act.api.helpers.handle_fact(
            actapi.fact("represents")
            .source("hash", results[myhash])
            .destination("content", content_id),
            output_format=output_format,
        )

    for cve in cves:
        act.api.helpers.handle_fact(
            actapi.fact("exploits")
            .source("content", content_id)
            .destination("vulnerability", cve),
            output_format=output_format,
        )

    for name, toolType in names:
        act.api.helpers.handle_fact(
            actapi.fact("classifiedAs")
            .source("content", content_id)
            .destination("tool", name),
            output_format=output_format,
        )

        # toolType may be None (as not all nameing schemes include toolType)
        if toolType:
            act.api.helpers.handle_fact(
                actapi.fact("classifiedAs")
                .source("tool", name)
                .destination("toolType", toolType),
                output_format=output_format,
            )

    if "detected_urls" in results:
        for u in map(
            urllib.parse.urlparse, [x["url"] for x in results["detected_urls"]]
        ):
            my_uri = add_uri(actapi, "fqdn", u.netloc, list(u))
            act.api.helpers.handle_fact(
                actapi.fact("at")
                .source("content", content_id)
                .destination("uri", my_uri),
                output_format=output_format,
            )
    if "undetected_urls" in results:
        for u in map(urllib.parse.urlparse, [x[0] for x in results["undetected_urls"]]):
            my_uri = add_uri(actapi, "fqdn", u.netloc, list(u))
            act.api.helpers.handle_fact(
                actapi.fact("at")
                .source("content", content_id)
                .destination("uri", my_uri),
                output_format=output_format,
            )


def handle_ip(
    actapi: act.api.Act, vtapi: VirusTotalApi, ip: Text, output_format: Text = "json"
) -> None:
    """Read IP address from stdin, query VirusTotal and
    output a JSON text readable by generic_uploaderr.py"""

    # To figure out what kind of IP address we have, let the ipaddress module
    # parse the string and test for instance type as the platform distinguishes
    # between IPv4 and IPv6 addresses.
    try:
        (ip_type, ip) = act.api.helpers.ip_obj(ip)
    except ValueError:
        return  # invalid address

    with no_ssl_verification():
        response = vtapi.get_ip_report(ip)

    try:
        results = response["results"]
    except KeyError:
        logging.error("%s in handle_ip for %s", response, ip)
        sys.exit(1)

    # create a dictionary of url that is observed in relation to the address.
    urls: collections.defaultdict = collections.defaultdict(list)
    if "detected_urls" in results:
        for u in map(
            urllib.parse.urlparse, [x["url"] for x in results["detected_urls"]]
        ):
            urls[u.netloc].append(u)
    if "undetected_urls" in results:
        for u in map(urllib.parse.urlparse, [x[0] for x in results["undetected_urls"]]):
            urls[u.netloc].append(u)

    if "resolutions" in results:
        for resolution in results["resolutions"]:
            if not resolution["hostname"]:
                warning("Empty hostname in resolution: ip=%s, result=%s", ip, results)
                continue
            act.api.helpers.handle_fact(
                actapi.fact("resolvesTo")
                .source("fqdn", resolution["hostname"])
                .destination(ip_type, ip),
                output_format=output_format,
            )
            # add all detected and undetected urls related to a given resolved hostname
            if resolution["hostname"] in urls:
                for u in urls[resolution["hostname"]]:
                    add_uri(actapi, "fqdn", resolution["hostname"], list(u))
    # if the actuall ip is part of the url, add the urls directly connected to the
    # ip.
    if ip in urls:
        for u in urls[ip]:
            add_uri(actapi, ip_type, ip, list(u))

    if "detected_downloaded_samples" in results:
        for sample in results["detected_downloaded_samples"]:

            my_uri = add_uri(actapi, ip_type, ip, ["network", ip, "", "", "", ""])

            act.api.helpers.handle_fact(
                actapi.fact("at")
                .source("content", sample["sha256"])
                .destination("uri", my_uri),
                output_format=output_format,
            )

            act.api.helpers.handle_fact(
                actapi.fact("represents")
                .source("hash", sample["sha256"])
                .destination("content", sample["sha256"]),
                output_format=output_format,
            )

            handle_hexdigest(
                actapi, vtapi, sample["sha256"], output_format=output_format
            )

    if "detected_communicating_samples" in results:
        for sample in results["detected_communicating_samples"]:
            my_uri = add_uri(actapi, ip_type, ip, ["network", ip, "", "", "", ""])

            act.api.helpers.handle_fact(
                actapi.fact("connectsTo")
                .source("content", sample["sha256"])
                .destination("uri", my_uri),
                output_format=output_format,
            )

            act.api.helpers.handle_fact(
                actapi.fact("represents")
                .source("hash", sample["sha256"])
                .destination("content", sample["sha256"]),
                output_format=output_format,
            )

            handle_hexdigest(
                actapi, vtapi, sample["sha256"], output_format=output_format
            )


def add_uri(
    actapi: act.api.Act,
    addr_type: str,
    addr: str,
    url: List[str],
    cache: dict = {},
    output_format: Text = "json",
) -> str:
    """Add a URI to the platform by creating the componentOf and scheme facts
    If called multiple times with arguments creating the same URI, the facts will only sent once.
    Return: The URI added"""
    try:
        my_uri = str(urllib.parse.urlunparse(url))
        if my_uri in cache:
            return my_uri

        cache[my_uri] = True

        act.api.helpers.handle_fact(
            actapi.fact("componentOf")
            .source(addr_type, addr)
            .destination("uri", my_uri),
            output_format=output_format,
        )

        act.api.helpers.handle_fact(
            actapi.fact("scheme", url[0]).source("uri", my_uri),
            output_format=output_format,
        )

        if url[2] and not url[2].strip() == "/":  # path
            act.api.helpers.handle_fact(
                actapi.fact("componentOf")
                .source("path", url[2])
                .destination("uri", my_uri),
                output_format=output_format,
            )

            basename = os.path.basename(url[2])
            if basename.strip():
                act.api.helpers.handle_fact(
                    actapi.fact("basename", basename).source("path", url[2]),
                    output_format=output_format,
                )

        if url[3]:  # query
            act.api.helpers.handle_fact(
                actapi.fact("componentOf")
                .source("query", url[3])
                .destination("uri", my_uri),
                output_format=output_format,
            )

    except act.api.base.ResponseError as e:
        error(str(e))

    return my_uri


def handle_domain(
    actapi: act.api.Act,
    vtapi: VirusTotalApi,
    domain: Text,
    output_format: Text = "json",
) -> None:
    """Read IP address from stdin, query VirusTotal and
    output a JSON text readable by generic_uploaderr.py"""

    with no_ssl_verification():
        response = vtapi.get_domain_report(domain)

    try:
        results = response["results"]
    except KeyError:
        logging.error("%s in handle_domain for %s", response, domain)
        sys.exit(1)

    if "detected_urls" in results:
        for u in map(
            urllib.parse.urlparse, [x["url"] for x in results["detected_urls"]]
        ):
            add_uri(actapi, "fqdn", domain, list(u))
    if "undetected_urls" in results:
        for u in map(urllib.parse.urlparse, [x[0] for x in results["undetected_urls"]]):
            add_uri(actapi, "fqdn", domain, list(u))

    if "resolutions" in results:
        for resolution in results["resolutions"]:
            ip = resolution["ip_address"]
            # To figure out what kind of IP address we have, let the ipaddress module
            # parse the string and test for instance type as the platform distinguishes
            # between IPv4 and IPv6 addresses.
            try:
                act.api.helpers.handle_fact(
                    actapi.fact("resolvesTo")
                    .source("fqdn", domain)
                    .destination(*act.api.helpers.ip_obj(ip)),
                    output_format=output_format,
                )
            except ValueError:
                continue  # invalid address

    if "detected_downloaded_samples" in results:
        for sample in results["detected_downloaded_samples"]:
            my_uri = add_uri(
                actapi, "fqdn", domain, ["network", domain, "", "", "", ""]
            )
            act.api.helpers.handle_fact(
                actapi.fact("at")
                .source("content", sample["sha256"])
                .destination("uri", my_uri),
                output_format=output_format,
            )
            act.api.helpers.handle_fact(
                actapi.fact("represents")
                .source("hash", sample["sha256"])
                .destination("content", sample["sha256"]),
                output_format=output_format,
            )
            handle_hexdigest(
                actapi, vtapi, sample["sha256"], output_format=output_format
            )

    if "detected_communicating_samples" in results:
        for sample in results["detected_communicating_samples"]:
            my_uri = add_uri(
                actapi, "fqdn", domain, ["network", domain, "", "", "", ""]
            )
            act.api.helpers.handle_fact(
                actapi.fact("connectsTo")
                .source("content", sample["sha256"])
                .destination("uri", my_uri),
                output_format=output_format,
            )
            act.api.helpers.handle_fact(
                actapi.fact("represents")
                .source("hash", sample["sha256"])
                .destination("content", sample["sha256"]),
                output_format=output_format,
            )
            handle_hexdigest(
                actapi, vtapi, sample["sha256"], output_format=output_format
            )


def handle_ioc(
    actapi: act.api.Act, vtapi: VirusTotalApi, ioc: Text, output_format: Text = "json"
) -> None:
    "Autodetect IOC type and send to correct handler"

    # Attempt to parse ioc AS IP address
    try:
        ipaddress.ip_address(ioc)
        info("Autodetected IOC as IP address: {}".format(ioc))
        handle_ip(actapi, vtapi, ioc, output_format=output_format)
        return
    except ValueError:  # Not IP, continue to next
        pass

    # md5, sha1 or sha256 or sha512 hash?
    if HASH_RE.search(ioc):
        info("Autodetected IOC as hash: {}".format(ioc))
        handle_hexdigest(actapi, vtapi, ioc, output_format=output_format)
        return

    # Assume domain
    info("Autodetection assumes IOC is domain: {}".format(ioc))
    handle_domain(actapi, vtapi, ioc, output_format=output_format)


def main() -> None:
    """main function"""

    # Look for default ini file in "/etc/actworkers.ini" and ~/config/actworkers/actworkers.ini
    # (or replace .config with $XDG_CONFIG_DIR if set)
    args = cli.handle_args(parseargs())

    actapi = worker.init_act(args)

    if not args.apikey:
        cli.fatal("You must specify --apikey on command line or in config file")

    in_data = sys.stdin.read().strip()

    proxies = (
        {"http": args.proxy_string, "https": args.proxy_string}
        if args.proxy_string
        else None
    )

    vtapi = VirusTotalApi(args.apikey, proxies=proxies)

    if args.hexdigest:
        handle_hexdigest(actapi, vtapi, in_data, output_format=args.output_format)

    elif args.ip:
        handle_ip(actapi, vtapi, in_data, output_format=args.output_format)

    elif args.domain:
        handle_domain(actapi, vtapi, in_data, output_format=args.output_format)

    else:  # Type not specified, autodetect
        handle_ioc(actapi, vtapi, in_data, output_format=args.output_format)


@contextlib.contextmanager
def no_ssl_verification() -> Generator[None, None, None]:
    """Monkey patch request to default to no verification of ssl"""

    old_request = requests.Session.request
    requests.Session.request = partialmethod(old_request, verify=False)  # type: ignore

    warnings.filterwarnings("ignore", "Unverified HTTPS request")
    yield
    warnings.resetwarnings()

    requests.Session.request = old_request  # type: ignore


def main_log_error() -> None:
    try:
        main()
    except Exception:
        error("Unhandled exception: {}".format(traceback.format_exc()))
        raise


if __name__ == "__main__":
    main_log_error()



"""Library for working with MISP export files"""

import enum
import ipaddress
import json
import re
import time
import uuid
from logging import warning
from typing import Callable, Dict, Optional, Text, Tuple


class ThreatLevelID(enum.Enum):
    """2.2.1.5.  threat_level_id

    threat_level_id represents the threat level.

    4:   Undefined
    3:   Low
    2:   Medium
    1:   High

    If a higher granularity is required, a MISP taxonomy applied as a Tag
    SHOULD be preferred.

    threat_level_id SHALL be present."""

    HIGH = 1
    MEDIUM = 2
    LOW = 3
    UNDEFINED = 4


class Analysis(enum.Enum):
    """2.2.1.6.  analysis

    analysis represents the analysis level.

    0:   Initial
    1:   Ongoing
    2:   Complete

    If a higher granularity is required, a MISP taxonomy applied as a Tag
    SHOULD be preferred.

    analysis SHALL be present."""

    INITIAL = 0
    ONGOING = 1
    COMPLETE = 2


class Distribution(enum.Enum):
    """2.2.1.13.  distribution

    distribution represents the basic distribution rules of the event.  The
    system must adhere to the distribution setting for access control and for
    dissemination of the event.

    distribution MUST be present and be one of the following options:

    0 Your Organisation Only
    1 This Community Only
    2 Connected Communities
    3 All Communities
    4 Sharing Group"""

    ORGANIZATION_ONLY = 0
    COMMUNITY_ONLY = 1
    CONNECTED_COMMUNITIES = 2
    ALL_COMMUNITIES = 3
    SHARING_GROUP = 4
    INHERIT_EVENT = 5


class Event:  # pylint: disable=R0902
    """Event represents a MISP event"""

    # --- MUST ----
    published = None
    info = None
    threat_level_id = ThreatLevelID.UNDEFINED
    analysis = Analysis.INITIAL
    date = None  # ISO-8601 (date only: YYYY-MM-DD) reference date of the event
    org_id = None  # Human readable org. generating the Event
    orgc_id = None  # Human readble org. *creating* the Event
    attribute_count = 0
    distribution = Distribution.ORGANIZATION_ONLY
    sharing_group_id = None
    # --- END_MUST ----

    # --- SHOULD ---
    extends_uuid = None
    # --- END_SHOULD ---

    tlp = None
    misp_galaxy: Dict[Text, Text] = {}

    def __init__(self, loads: Optional[Text] = None) -> None:
        self._uuid: uuid.UUID = uuid.uuid4()
        self.timestamp: int = int(
            time.time()
        )  # Timestamp of event creation or event/attribute last update
        self.publish_timestamp: int = int(
            time.time()
        )  # reference time when the event was published on the instance.

        if loads:
            data = json.loads(loads)
            event = data["Event"]
            self._uuid = uuid.UUID(event["uuid"])
            self.published = event["published"]
            self.info = event["info"]
            self.threat_level_id = ThreatLevelID(int(event["threat_level_id"]))
            self.analysis = Analysis(int(event["analysis"]))
            self.date = event["date"]
            self.timestamp = event["timestamp"]

            if "publish_timestamp" in event:
                self.publish_timestamp = event["publish_timestamp"]
            else:
                warning("publish_timestamp missing in event %s", event)
            self.org_id = event.get("org_id", "N/A")
            self.orgc_id = event.get("orgc_id", "N/A")
            self.attribute_count = event.get("attribute_count", 0)
            self.distribution = Distribution(int(event.get("distribution", 0)))
            self.sharing_group_id = event.get("sharing_group_id", None)
            self.extends_uuid = event.get("extends_uuid", None)

            misp_re = re.compile(r'misp-galaxy:(.*?)="(.*?)"')
            for tag in event.get("Tag", []):
                name = tag["name"]
                if name.startswith("tlp"):
                    self.tlp = name.split(":")[1]
                if name.startswith("misp-galaxy"):
                    for match in misp_re.findall(name):
                        self.misp_galaxy[match[0]] = match[1]

            self.attributes = [Attribute(e) for e in event.get("Attribute", [])]
            objects = event.get("Object", [])
            for obj in objects:
                obj_attributes = obj.get("Attribute", [])
                self.attributes += [Attribute(e) for e in obj_attributes]

    def __str__(self) -> Text:
        return "({0}) {1} - {2} ".format(self.timestamp, self._uuid, self.info)

    @property
    def uuid(self):  # type: ignore
        """The Event UUID property"""
        return self._uuid


class Attribute:  # attributeattributes in misp babel pylint: disable=R0903
    """Represents a MISP attribute, performing the neccessary translations between
    misp and act namespaces"""

    def __init__(self, attributedict: Dict[Text, Text]):

        try:
            self._uuid = attributedict["uuid"]
            self.id: Text = attributedict["uuid"]
            mapper_fn = map_misp_to_act.get(
                attributedict["type"], lambda x: (None, None)
            )
            self.act_type: Optional[Text] = None
            self.value: Optional[Text] = None

            self.act_type, self.value = mapper_fn(attributedict["value"])

            if (
                "RelatedAttribute" in attributedict
                and attributedict["RelatedAttribute"]
            ):
                print("DEBUG: {0}".format(attributedict["RelatedAttribute"]))
        except:  # noqa=E722 (silence Flake8; catching for logging and re-raise)
            print(attributedict)
            print(attributedict["value"][:100])
            raise

    def __str__(self) -> Text:
        return "{0} {1}:{2}".format(self.id, self.act_type, self.value)


def hash_f(x: Text) -> Tuple[Text, Text]:
    """Translation function misp -> act"""
    return "hash", x.lower()


def certificate_f(x: Text) -> Tuple[Text, Text]:
    """Translation function misp -> act"""
    return "certificate", x.lower()


def threat_actor_f(x: Text) -> Tuple[Text, Text]:
    """Translation function misp -> act"""
    return "threatActor", x


def campaign_f(x: Text) -> Tuple[Text, Text]:
    """Translation function misp -> act"""
    return "campaign", x.lower()


def email_f(x: Text) -> Tuple[Text, Text]:
    """Translation function misp -> act"""
    return "uri", "email://{}".format(x.lower())


def person_f(x: Text) -> Tuple[Text, Text]:
    """Translation function misp -> act"""
    return "person", x.lower()


def organization_f(x: Text) -> Tuple[Text, Text]:
    """Translation function misp -> act"""
    return "organization", x.lower()


def fqdn_f(x: Text) -> Tuple[Text, Text]:
    """Translation function misp -> act"""
    return "fqdn", x.lower()


def ip_f(x: Text) -> Tuple[Optional[Text], Optional[Text]]:
    """Translation function misp -> act"""
    try:
        addrv6 = ipaddress.IPv6Address(x)
        return "ipv6", str(addrv6.exploded)
    except ipaddress.AddressValueError:
        try:
            ipaddress.IPv4Address(x)
            return "ipv4", x
        except ipaddress.AddressValueError:
            pass

    return None, None


def uri_f(x: Text) -> Tuple[Text, Text]:
    """Translation function misp -> act"""
    if not x.startswith("http"):
        x = "http://{0}".format(x)
    return "uri", x


def user_agent_f(x: Text) -> Tuple[Text, Text]:
    """Translation function misp -> act"""
    return "userAgent", x


def vulnerability_f(x: Text) -> Tuple[Text, Text]:
    """Translation function misp -> act"""
    return "vulnerability", x.lower()


def mutex_f(x: Text) -> Tuple[Text, Text]:
    """Translation function misp -> act"""
    return "mutex", x


map_misp_to_act: Dict[Text, Callable[[Text], Tuple[Optional[Text], Optional[Text]]]] = {
    "authentihash": hash_f,
    "campaign-name": campaign_f,
    "hostname": fqdn_f,
    "domain": fqdn_f,
    "impfuzzy": hash_f,
    "imphash": hash_f,
    "ip-dst": ip_f,
    "ip-src": ip_f,
    "link": uri_f,
    "md5": hash_f,
    "mutex": mutex_f,
    "sha1": hash_f,
    "sha224": hash_f,
    "sha256": hash_f,
    "sha384": hash_f,
    "sha512/224": hash_f,
    "sha512/256": hash_f,
    "sha512": hash_f,
    "ssdeep": hash_f,
    "threat-actor": threat_actor_f,
    "url": uri_f,
    "user-agent": user_agent_f,
    "vulnerability": vulnerability_f,
    "whois-registrant-email": email_f,
    "whois-registrant-name": person_f,
    "whois-registrar": organization_f,
    "x509-fingerprint-sha1": certificate_f,
}



#!/usr/bin/env python3.6
"""MISP feed worker pulling down feeds in misp_feeds.txt
and adding data to the platform"""

import argparse
import hashlib
import json
import os
import sys
import traceback
from logging import error, info
from typing import Dict, Generator, Optional, Text

import caep
import requests

import act
import act.api.helpers
from act.workers.libs import misp, worker
from act.api.libs import cli

try:
    import urlparse
except ModuleNotFoundError:  # Python3
    import urllib.parse as urlparse  # type: ignore


def parseargs() -> argparse.ArgumentParser:
    """Parse arguments"""
    parser = worker.parseargs("Get MISP feeds from MISP sharing directories")

    parser.add_argument(
        "--manifest-dir",
        default=caep.get_cache_dir("misp_manifest"),
        help="The directory to store latest manifests",
    )

    return parser


def verify_manifest_dir(manifest_dir: Text) -> None:
    """Verify that the directory structure exists and that there is
    always a feed file (Even empty)"""

    # Manifest is at default location - create directory if it does not exists
    if manifest_dir == caep.get_cache_dir("misp_manifest"):
        caep.get_cache_dir("misp_manifest", create=True)

    # If there is specified a manifest directory in the .ini file we
    # verify that it exists (or fail hard). If no such directory
    # is defined, we default to using $XDG_CACHE_DIR and create a new
    # 'misp_maifest' directory there.
    if not os.path.isdir(manifest_dir):
        print("Could not open manifest directory:", manifest_dir)
        sys.exit(1)

    # Check that the misp_feeds.txt file actually exists. If not 'touch'
    # the file to make sure there is at least some default config present.
    feed_file = os.path.join(manifest_dir, "misp_feeds.txt")
    if not os.path.isfile(feed_file):
        with open(feed_file, "w") as feed_h:
            feed_h.write("https://www.circl.lu/doc/misp/feed-osint/")


def handle_event_file(
    feed_url: Text,
    uuid: Text,
    proxy_string: Optional[Text] = None,
    cert_file: Optional[Text] = None,
) -> misp.Event:
    """Download, parse and store single event file"""

    info("Handling {0} from {1}".format(uuid, feed_url))

    proxies: Optional[Dict[Text, Text]] = None

    if proxy_string:
        proxies = {"http": proxy_string, "https": proxy_string}

    url = urlparse.urljoin(feed_url, "{0}.json".format(uuid))
    req = requests.get(url, proxies=proxies, verify=cert_file)
    return misp.Event(loads=req.text)


def handle_feed(
    manifest_dir: Text,
    feed_url: Text,
    proxy_string: Optional[Text] = None,
    cert_file: Optional[Text] = None,
) -> Generator[misp.Event, None, None]:
    """Get the manifest file, check if an event file is downloaded
    before (cache) and dispatch event handling of separate files"""

    proxies: Optional[Dict[Text, Text]] = None

    if proxy_string:
        proxies = {"http": proxy_string, "https": proxy_string}

    manifest_url = urlparse.urljoin(feed_url, "manifest.json")
    req = requests.get(manifest_url, proxies=proxies, verify=cert_file)

    manifest = json.loads(req.text)

    feed_sha1 = hashlib.sha1(feed_url.encode("utf-8")).hexdigest()

    old_manifest = {}
    if manifest_dir != "NODIR":
        try:
            with open(os.path.join(manifest_dir, feed_sha1)) as infile:
                old_manifest = json.load(infile)
        except IOError:
            pass

    for uuid in manifest:
        if uuid not in old_manifest:
            yield handle_event_file(feed_url, uuid, proxy_string, cert_file)

    if manifest_dir != "NODIR":
        with open(os.path.join(manifest_dir, feed_sha1), "wb") as outfile:
            outfile.write(json.dumps(manifest).encode("utf-8"))


def main() -> None:
    """program entry point"""

    # Look for default ini file in "/etc/actworkers.ini" and ~/config/actworkers/actworkers.ini
    # (or replace .config with $XDG_CONFIG_DIR if set)
    args = cli.handle_args(parseargs())

    manifest_dir = args.manifest_dir

    actapi = worker.init_act(args)

    verify_manifest_dir(manifest_dir)
    misp_feeds_file = os.path.join(manifest_dir, "misp_feeds.txt")

    with open(misp_feeds_file) as f:
        for line in f:
            feed_data = handle_feed(
                manifest_dir, line.strip(), args.proxy_string, args.cert_file
            )
            for event in feed_data:
                n = 0
                e = 0

                act.api.helpers.handle_fact(
                    actapi.fact("name", event.info).source("report", str(event.uuid)),
                    output_format=args.output_format,
                )

                n += 1

                try:
                    act.api.helpers.handle_fact(
                        actapi.fact("externalLink")
                        .source("uri", "{0}/{1}.json".format(line.strip(), event.uuid))
                        .destination("report", str(event.uuid)),
                        output_format=args.output_format,
                    )

                    n += 1
                except act.api.base.ResponseError as err:
                    e += 1
                    error(
                        "misp_feeds, main unable to add fact to platform, error calling actapi: %s"
                        % err,
                        exc_info=True,
                    )

                for attribute in event.attributes:
                    if not attribute.act_type:
                        continue
                    try:
                        act.api.helpers.handle_fact(
                            actapi.fact("mentions")
                            .source("report", str(event.uuid))
                            .destination(attribute.act_type, attribute.value),
                            output_format=args.output_format,
                        )
                        n += 1
                    except act.api.base.ResponseError as err:
                        e += 1
                        error(
                            "misp_feeds: main unable to add attribute fact to platform, error calling actapi: %s"
                            % err,
                            exc_info=True,
                        )
                info("{0} facts. {1} errors.".format(n, e))


def main_log_error() -> None:
    "Call main() and log all exceptions as errors"
    try:
        main()
    except Exception:
        error("Unhandled exception: {}".format(traceback.format_exc()))
        raise


if __name__ == "__main__":
    main_log_error()



import pytest
from act.workers.libs import misp
from act.workers import misp_feeds
import json
import uuid


@pytest.mark.server(
    url="/manifest.json",
    response=json.load(open("test/data/misp_dir/manifest.json")),
    method="GET",
)
@pytest.mark.server(
    url="/57e27c3d-17a8-4c3b-8195-493e950d210f.json",
    response=json.load(
        open("test/data/misp_dir/57e27c3d-17a8-4c3b-8195-493e950d210f.json")
    ),
    method="GET",
)
@pytest.mark.server(
    url="/5721c2e4-05ec-4af3-9264-411b950d210f.json",
    response=json.load(
        open("test/data/misp_dir/5721c2e4-05ec-4af3-9264-411b950d210f.json")
    ),
    method="GET",
)
def test_event() -> None:
    info = [
        "Malspam (2016-04-28) - Locky (#2)",
        'Malspam 2016-09-21 (.wsf in .zip) - campaign: "E-TICKET {integer}"',
    ]
    uuids = [
        "5721c2e4-05ec-4af3-9264-411b950d210f",
        "57e27c3d-17a8-4c3b-8195-493e950d210f",
    ]
    dates = ["2016-04-28", "2016-09-21"]

    for event in misp_feeds.handle_feed("NODIR", "http://localhost:5000"):
        assert event.published
        assert str(event.uuid) in uuids
        assert event.date in dates
        assert event.threat_level_id == misp.ThreatLevelID.LOW
        assert not event.extends_uuid
        assert event.analysis in [misp.Analysis.INITIAL, misp.Analysis.COMPLETE]
        assert event.info in info


def test_load_event() -> None:
    with open("test/misp_event.json") as event_f:
        a = misp.Event(loads=event_f.read())
    assert (
        a.info
        == 'M2M -  Dridex 2017-11-14 : botnet 7200 : "Invoice No.\n 123456" - "Invoice-123456-06.doc"'
    )
    assert a.uuid == uuid.UUID("5a0f0fb1-0b54-4ace-bb7b-429f950d210f")
    assert a.date == "2017-11-17"
    assert a.published is False
    assert a.threat_level_id == misp.ThreatLevelID.LOW
    assert a.analysis == misp.Analysis.ONGOING
    assert "tool" in a.misp_galaxy
    assert a.misp_galaxy.get("tool", "") == "Dridex"



"""Common worker library"""

import argparse
import json
import smtplib
import socket
import sys
import urllib.parse
from email.mime.text import MIMEText
from logging import warning
from typing import Any, Optional

import act.api
import requests
import urllib3
from act.types.format import object_format
from act.types.types import object_validates


class NotAllowed(Exception):
    """UnsupportedScheme is used when parsing URLs that does not contain a supported scheme"""

    def __init__(self, *args: Any) -> None:
        Exception.__init__(self, *args)


class FetchError(Exception):
    """UnsupportedScheme is used when parsing URLs that does not contain a supported scheme"""

    def __init__(self, *args: Any) -> None:
        Exception.__init__(self, *args)


class UnsupportedScheme(Exception):
    """UnsupportedScheme is used when parsing URLs that does not contain a supported scheme"""

    def __init__(self, *args: Any) -> None:
        Exception.__init__(self, *args)


class NoResult(Exception):
    """NoResult is used in API request (no data returned)"""

    def __init__(self, *args: Any) -> None:
        Exception.__init__(self, *args)


class UnknownFormat(Exception):
    """UnknownFormat is used on unknown parsing formats"""

    def __init__(self, *args: Any) -> None:
        Exception.__init__(self, *args)


class ResourceLimitExceeded(Exception):
    """Resource Limits Exceeded"""

    def __init__(self, *args: Any) -> None:
        Exception.__init__(self, *args)


class ServiceTimeout(Exception):
    """Internal service timeouts"""

    def __init__(self, *args: Any) -> None:
        Exception.__init__(self, *args)


class AuthenticationError(Exception):
    """Authentication Errors"""

    def __init__(self, *args: Any) -> None:
        Exception.__init__(self, *args)


def parseargs(description: str) -> argparse.ArgumentParser:
    """Parse arguments"""

    parser = act.api.libs.cli.parseargs(description, fact_arguments=True)

    parser.add_argument(
        "--disabled",
        dest="disabled",
        action="store_true",
        help="Worker is disabled (exit immediately)",
    )
    return parser


def init_act(
    args: argparse.Namespace, no_format: bool = False, no_validate: bool = False
) -> act.api.Act:

    """Initialize act api from arguments"""

    api = act.api.libs.cli.init_act(
        args,
        object_formatter=None if no_format else object_format,
        object_validator=None if no_validate else object_validates,
    )

    # This check is done here to make sure logging is set up
    if args.disabled:
        warning("Worker is disabled")
        sys.exit(0)

    return api


def fetch(
    url: str, proxy_string: Optional[str], timeout: int = 60, verify_https: bool = False
) -> Any:
    """Fetch remote URL and return content
    url (string):                    File or URL to fetch
    proxy_string (string, optional): Optional proxy string on format host:port
    timeout (int, optional):         Timeout value for query (default=60 seconds)
    """

    proxies = {"http": proxy_string, "https": proxy_string}

    options = {
        "verify": verify_https,
        "timeout": timeout,
        "proxies": proxies,
        "params": {},
    }

    parsed = urllib.parse.urlparse(url)

    # No scheme - assume this is a file
    if not parsed.scheme:
        return open(url).read()

    if not parsed.scheme.lower() in ("http", "https"):
        raise UnsupportedScheme(f"Unsupported scheme in {url}")

    if not verify_https:
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

    try:
        req = requests.get(url, **options)
    except (
        urllib3.exceptions.ReadTimeoutError,
        requests.exceptions.ReadTimeout,
        socket.timeout,
    ) as err:
        raise FetchError(
            "Timeout ({0.__class__.__name__}), query: {1}".format(err, url)
        )

    if not req.status_code == 200:
        errmsg = "status_code: {0.status_code}: {0.content}"
        raise FetchError(errmsg.format(req))

    return req.text


def fetch_json(
    url: str, proxy_string: Optional[str], timeout: int = 60, verify_https: bool = False
) -> Any:
    """Fetch remote URL or local and return content parse as json
    url (string):                    File or URL to fetch
    proxy_string (string, optional): Optional proxy string on format host:port
    timeout (int, optional):         Timeout value for query (default=60 seconds)
    """
    content = fetch(url, proxy_string, timeout, verify_https)
    try:
        return json.loads(content)
    except json.decoder.JSONDecodeError as e:
        raise FetchError(f"Cannot parse as json {e}, {content} {url}")


def sendmail(
    smtphost: str, sender: str, recipient: str, subject: str, body: str
) -> None:
    """Send email"""

    msg = MIMEText(body, "plain", "utf-8")
    msg["Subject"] = subject
    msg["From"] = sender
    msg["To"] = recipient
    s = smtplib.SMTP(smtphost)
    s.sendmail(sender, [recipient], msg.as_string())
    s.quit()



#!/usr/bin/env python3.6
"""

act feed worker pulling down fact bundles and either:
- dumping to directory
or
- upload to platform

"""

import argparse
import gzip
import hashlib
import io
import os
import traceback
import urllib.parse as urlparse
from logging import error, info, warning
from pathlib import Path
from typing import Dict, Optional, Text

import act.api
import act.api.helpers
import caep
import pid
import requests
from act.api.libs import cli

from act.workers.generic_uploader import uploader
from act.workers.libs import worker


def parseargs() -> argparse.ArgumentParser:
    """Parse arguments"""
    parser = worker.parseargs("Get act feeds from act sharing directories")

    parser.add_argument(
        "--feed-uri",
        help="URI to retrieve feed from",
    )

    parser.add_argument(
        "--dump-dir",
        type=Path,
        help="Dump manifest/bundles to directory instead of sendings facts to uploader",
    )

    parser.add_argument(
        "--feed-cache",
        type=Path,
        default=caep.get_cache_dir("act_feed_cache"),
        help="The directory to store information about last run",
    )

    parser.add_argument(
        "--no-exit-on-error",
        action="store_true",
        help="Log errors and continue on platform upload errors",
    )

    return parser


def get_last_run_filename(feed_cache: Path, feed_uri: Text) -> Path:
    """Use hash of feed as cache file, so you can run multiple processes with feed"""
    if str(feed_cache) == caep.get_cache_dir("act_feed_cache"):
        caep.get_cache_dir("act_feed_cache", create=True)

    return feed_cache / Path(hashlib.sha256(feed_uri.encode()).hexdigest())


def get_last_run(last_run_file: Path) -> int:
    """Get last run"""

    if not last_run_file.is_file():
        warning("Last run file not found, assuming first run")
        return 0

    with open(last_run_file) as f:
        return int(f.read())


def update_last_run(last_run_file: Path, last_run: int) -> None:
    """Update last run file"""

    with open(last_run_file, "w") as f:
        f.write(str(last_run))


def handle_bundle(
    feed_uri: Text,
    bundle: Text,
    actapi: act.api.Act,
    dump_dir: Path,
    proxies: Optional[Dict[Text, Text]] = None,
    cert_file: Optional[Text] = None,
    no_exit_on_error: bool = False,
) -> None:
    """Retrieve and handle bundle file"""

    bundle_uri = urlparse.urljoin(feed_uri, bundle)

    req = requests.get(bundle_uri, proxies=proxies, verify=cert_file)

    if dump_dir:
        info("Storing to %s, %s", dump_dir, bundle_uri)
        with open(dump_dir / Path(bundle), "wb") as f:
            f.write(req.content)

    else:
        info("Iterate over facts in uploader: %s", bundle_uri)

        # uploader will print facts to stdout unless act_baseurl is set
        with gzip.open(io.BytesIO(req.content), "rb") as facts:
            uploader(actapi, facts, no_exit_on_error=no_exit_on_error)


def handle_feed(
    last_run_filename: Path,
    feed_uri: Text,
    actapi: act.api.Act,
    dump_dir: Path,
    proxies: Optional[Dict[Text, Text]] = None,
    cert_file: Optional[Text] = None,
    no_exit_on_error: bool = False,
) -> int:
    """Get the manifest file, and handle new bundles"""

    last_run = get_last_run(last_run_filename)

    # Ensure URI ends with "/" so it can be joined with manifests/bundles
    if not feed_uri.endswith("/"):
        feed_uri += "/"

    manifest_uri = urlparse.urljoin(feed_uri, "manifest.json")
    req = requests.get(manifest_uri, proxies=proxies, verify=cert_file)

    manifest = req.json()

    if dump_dir:
        with open(dump_dir / Path("manifest.json"), "w") as f:
            f.write(req.text)

    # for each bundle handled, the update time will be stored to disk
    # so it needs to be sorted by update time
    manifests = sorted(manifest["bundles"].items(), key=lambda item: item[1])

    for bundle, updated in manifests:
        # Skip if the file is modified *before* last run
        if updated <= last_run:
            info(
                "Skipping bundle %s, updated %s, last_run=%s", bundle, updated, last_run
            )
            continue

        handle_bundle(
            feed_uri,
            bundle,
            actapi,
            dump_dir,
            proxies,
            cert_file,
            no_exit_on_error,
        )

        # update last run after sucessfull handling of bundle file
        update_last_run(last_run_filename, updated)


def main() -> None:
    """program entry point"""

    # Look for default ini file in "/etc/actworkers.ini" and
    # ~/config/actworkers/actworkers.ini
    # (or replace .config with $XDG_CONFIG_DIR if set)
    args = cli.handle_args(parseargs())

    if not args.feed_uri:
        cli.fatal("--feed-uri not specified")

    if args.dump_dir and not args.dump_dir.is_dir():
        os.makedirs(args.dump_dir)

    proxies = (
        {"http": args.proxy_string, "https": args.proxy_string}
        if args.proxy_string
        else None
    )

    actapi = worker.init_act(args)

    try:
        # Get "updated" from last successful run
        last_run_filename = get_last_run_filename(args.feed_cache, args.feed_uri)

        with pid.PidFile(force_tmpdir=True, pidname="act_feed.pid"):
            handle_feed(
                last_run_filename,
                args.feed_uri,
                actapi,
                args.dump_dir,
                proxies,
                args.cert_file,
                args.no_exit_on_error,
            )

    except pid.base.PidFileAlreadyLockedError:
        error("pid file found - feed is already running")


def main_log_error() -> None:
    "Call main() and log all exceptions as errors"
    try:
        main()
    except Exception:
        error("Unhandled exception: {}".format(traceback.format_exc()))
        raise


if __name__ == "__main__":
    main_log_error()



""" Test for scio worker """
import json

import act.api
from act.types.format import object_format
from act.types.types import object_validates

from act.workers import scio


def test_scio2_facts(capsys) -> None:  # type: ignore
    """Test for scio2 facts, by comparing to captue of stdout"""
    with open("test/scio-doc.json") as scio_doc:
        doc = json.loads(scio_doc.read())

    api = act.api.Act(
        "",
        None,
        "error",
        strict_validator=True,
        object_formatter=object_format,
        object_validator=object_validates,
    )
    act.api.helpers.handle_fact.cache_clear()

    scio.add_to_act(api, doc, output_format="str")

    captured = capsys.readouterr()

    facts = set(captured.out.split("\n"))

    report_id = doc["hexdigest"]

    sha256 = doc["indicators"]["sha256"][0]
    uri = doc["indicators"]["uri"][0]  # "http://www.us-cert.gov/tlp."

    fact_assertions = [
        api.fact("name", "TA18-149A.stix.xml").source("report", report_id),
        api.fact("mentions")
        .source("report", report_id)
        .destination("ipv4", "187.127.112.60"),
        api.fact("mentions")
        .source("report", report_id)
        .destination("ipv6", "0000:0000:0000:0000:0000:0000:0000:0001"),
        api.fact("mentions")
        .source("report", report_id)
        .destination("hash", "4613f51087f01715bf9132c704aea2c2"),
        api.fact("mentions").source("report", report_id).destination("hash", sha256),
        api.fact("mentions")
        .source("report", report_id)
        .destination("country", "Colombia"),
        api.fact("mentions").source("report", report_id).destination("uri", uri),
        api.fact("represents")
        .source("report", report_id)
        .destination("content", report_id),
        api.fact("at").source("content", report_id).destination("uri", doc["uri"]),
        api.fact("componentOf")
        .source("fqdn", "www.us-cert.gov")
        .destination("uri", uri),
        api.fact("componentOf").source("path", "/tlp.").destination("uri", uri),
        api.fact("scheme", "http").source("uri", uri),
        api.fact("mentions").source("report", report_id).destination("tool", "cobra"),
        api.fact("mentions")
        .source("report", report_id)
        .destination("threatActor", "hidden cobra"),
        api.fact("mentions")
        .source("report", report_id)
        .destination("sector", "finance"),
        api.fact("mentions")
        .source("report", report_id)
        .destination("uri", "email://redhat@gmail.com"),
        api.fact("mentions")
        .source("report", report_id)
        .destination("ipv4Network", "192.168.0.0/16"),
        api.fact("represents").source("hash", sha256).destination("content", sha256),
        api.fact("mentions")
        .source("report", report_id)
        .destination("vulnerability", "cve-2019-222"),
        api.fact("mentions")
        .source("report", report_id)
        .destination("vulnerability", "ms16-034"),
        api.fact("mentions")
        .source("report", report_id)
        .destination("technique", "T1055"),
        api.fact("mentions")
        .source("report", report_id)
        .destination("technique", "T1547.001"),
        api.fact("mentions")
        .source("report", report_id)
        .destination("tactic", "TA0003"),
    ]

    for fact_assertion in fact_assertions:
        if not str(fact_assertion) in facts:
            print(f"{fact_assertion} is missing")

        assert str(fact_assertion) in facts



#!/usr/bin/env python3

"""
Threat Actor Helper worker

Create common placeholders based on known information.

"""

import argparse
import functools
import sys
import traceback
from logging import error
from typing import Dict, List, Text

import act.api
from act.api.helpers import Act, handle_facts
from pyattck import Attck

from act.workers import worker_config
from act.workers.libs import worker
from act.api.libs import cli


class VocabularyException(Exception):
    pass


def parseargs() -> argparse.ArgumentParser:
    """Parse arguments"""
    parser = worker.parseargs("Threat Actor helper")
    parser.add_argument("--ta", "--threat-actor", help="Threat Actor")
    parser.add_argument("--ta-located-in", help="Threat Actor located in")
    parser.add_argument("--campaign", help="Campaign")
    parser.add_argument(
        "--techniques",
        type=worker_config.string_list,
        default=[],
        help="Techniques (Commaseparated list)",
    )
    parser.add_argument(
        "--tools",
        type=worker_config.string_list,
        default=[],
        help="Tools (Commaseparated list)",
    )
    parser.add_argument(
        "--sectors",
        type=worker_config.string_list,
        default=[],
        help="Sectors (Commaseparated list)",
    )
    parser.add_argument(
        "--target-countries",
        type=worker_config.string_list,
        default=[],
        help="Target Countries (Commaseparated list)",
    )
    parser.add_argument(
        "--country-region",
        default="https://raw.githubusercontent.com/lukes/ISO-3166-Countries-with-Regional-Codes/master/all/all.json",
        help="Country region in json format (HTTP URL or file)",
    )

    # Stix Vocabulary
    parser.add_argument(
        "--sector-vocabulary",
        default="https://raw.githubusercontent.com/mnemonic-no/act-scio2/master/act/scio/etc/plugins/sectors.cfg",
        help="Sector vocabulary (STIX2, in scio format). Fetched from URL or file",
    )

    return parser


def levenshtein(a: Text, b: Text) -> int:
    "Calculates the Levenshtein distance between a and b."

    n, m = len(a), len(b)
    if n > m:
        # Make sure n <= m, to use O(min(n,m)) space
        a, b = b, a
        n, m = m, n

    current = list(range(n + 1))
    for i in range(1, m + 1):
        previous, current = current, [i] + [0] * n
        for j in range(1, n + 1):
            add, delete = previous[j] + 1, current[j - 1] + 1
            change = previous[j - 1]
            if a[j - 1] != b[i - 1]:
                change = change + 1
            current[j] = min(add, delete, change)

    return current[n]


def find_levenshtein_candiates(value: Text, vocabulary: List, n_items: int = 3) -> List:
    "Return top N list of possible levenshtein candidates"

    possible_match = []

    for entry in set(vocabulary):
        possible_match.append([levenshtein(value, entry), entry])

    return [x[1] for x in sorted(possible_match)[:n_items] if x[1]]


@functools.lru_cache(32)
def fetch_country_regions(
    country_region: Text, proxy_string: Text, http_timeout: int
) -> Dict[Text, Text]:
    "Fetch ISO-3166 list of country/regions"

    vocabulary: Dict = {}

    for c_map in worker.fetch_json(country_region, proxy_string, http_timeout):
        country = c_map["name"]
        vocabulary[country.lower()] = country

    return vocabulary


@functools.lru_cache(32)
def fetch_sector(
    sector_vocabulary: Text, proxy_string: Text, http_timeout: int
) -> Dict[Text, Text]:
    "Fetch STIX2 sector vocabulary"

    vocabulary: Dict = {}

    # format=
    # name: alias1,alias2,...aliasN
    for line in worker.fetch(sector_vocabulary, proxy_string, http_timeout).split("\n"):
        if not line.strip():
            continue

        name, aliases = line.split(":", 1)

        vocabulary[name] = name

        for alias in [a.strip() for a in aliases.split(",")]:
            vocabulary[alias] = name

    return vocabulary


@functools.lru_cache(32)
def mitre_techniques() -> Dict[Text, Text]:
    """
    Get list of all MITRE ATT&CK techniques and return map combined with
    both technique IDs and (lowercased) name as key an name as value
    """

    attack = Attck()
    techniques = {}

    for technique in attack.enterprise.techniques:
        techniques[technique.id] = technique.name
        techniques[technique.name.lower()] = technique.name

    return techniques


def technique_lookup(technique: Text) -> Text:
    "Lookup technique and return name if exists, otherwise raise VocabularyException"
    techniques = mitre_techniques()

    try:
        return techniques[technique]
    except KeyError:
        raise VocabularyException(
            "Technique {} not found in MITRE ATT&CK. Nearest matches: {}".format(
                technique,
                ", ".join(find_levenshtein_candiates(technique, techniques.keys())),
            )
        )


def country_lookup(args: argparse.Namespace, country: Text) -> Text:
    "Lookup country and return name if exists, otherwise raise VocabularyException"
    vocabulary = fetch_country_regions(
        args.country_region, args.proxy_string, args.http_timeout
    )

    try:
        return vocabulary[country.lower()]
    except KeyError:
        raise VocabularyException(
            "Country {} not found in ISO 3166 database. Nearest matches: {}".format(
                country,
                ", ".join(find_levenshtein_candiates(country, vocabulary.values())),
            )
        )


def sector_lookup(args: argparse.Namespace, sector: Text) -> Text:
    "Lookup sector and return name if exists, otherwise raise VocabularyException"

    vocabulary = fetch_sector(
        args.sector_vocabulary, args.proxy_string, args.http_timeout
    )

    try:
        return vocabulary[sector.lower()]
    except KeyError:
        raise VocabularyException(
            "Sector {} not found in Stix vocabulary. Nearest matches: {}".format(
                sector, ", ".join(find_levenshtein_candiates(sector, vocabulary))
            )
        )


def add_ta_techniques(
    client: Act, output_format: Text, threat_actor: Text, techniques: List[Text]
) -> None:
    """Threat Actor Techniques"""

    for technique in techniques:
        handle_facts(
            act.api.fact.fact_chain(
                client.fact("attributedTo")
                .source("incident", "*")
                .destination("threatActor", threat_actor),
                client.fact("observedIn")
                .source("technique", technique)
                .destination("incident", "*"),
            ), 
            output_format=output_format)


def add_ta_tools(
    client: Act, output_format: Text, threat_actor: Text, tools: List[Text]
) -> None:
    """Threat Actor Tools"""

    for tool in tools:
        handle_facts(
            act.api.fact.fact_chain(
                client.fact("classifiedAs")
                .source("content", "*")
                .destination("tool", tool),
                client.fact("observedIn")
                .source("content", "*")
                .destination("incident", "*"),
                client.fact("attributedTo")
                .source("incident", "*")
                .destination("threatActor", threat_actor),
            ), 
            output_format=output_format)


def add_ta_sectors(
    client: Act, output_format: Text, threat_actor: Text, sectors: List[Text]
) -> None:
    """Threat Actor Sectors"""
    for sector in sectors:
        handle_facts(
            act.api.fact.fact_chain(
                client.fact("targets")
                .source("incident", "*")
                .destination("organization", "*"),
                client.fact("memberOf")
                .source("organization", "*")
                .destination("sector", sector),
                client.fact("attributedTo")
                .source("incident", "*")
                .destination("threatActor", threat_actor),
            ), 
            output_format=output_format)


def add_ta_target_country(
    client: Act, output_format: Text, threat_actor: Text, target_countries: List[Text]
) -> None:
    """Threat actor target countries"""
    for target_country in target_countries:
        handle_facts(
            act.api.fact.fact_chain(
                client.fact("targets")
                .source("incident", "*")
                .destination("organization", "*"),
                client.fact("locatedIn")
                .source("organization", "*")
                .destination("country", target_country),
                client.fact("attributedTo")
                .source("incident", "*")
                .destination("threatActor", threat_actor),
        ), output_format=output_format)


def add_ta_located_in(
    client: Act, output_format: Text, threat_actor: Text, located_in: Text
) -> None:
    """Threat actor located in"""
    handle_facts(
        act.api.fact.fact_chain(
            client.fact("locatedIn")
            .source("organization", "*")
            .destination("country", located_in),
            client.fact("attributedTo")
            .source("threatActor", threat_actor)
            .destination("organization", "*"),
        ), output_format=output_format)


def add_ta_campaign(
    client: Act, output_format: Text, threat_actor: Text, campaign: Text
) -> None:
    """Threat Actor Campaign"""
    handle_facts(
        act.api.fact.fact_chain(
            client.fact("attributedTo")
            .source("incident", "*")
            .destination("campaign", campaign),
            client.fact("attributedTo")
            .source("incident", "*")
            .destination("threatActor", threat_actor),
        ), output_format=output_format)


def main() -> None:
    """Main function"""

    # Look for default ini file in "/etc/actworkers.ini" and ~/config/actworkers/actworkers.ini
    # (or replace .config with $XDG_CONFIG_DIR if set)
    args = cli.handle_args(parseargs())

    actapi = worker.init_act(args)

    if not args.ta:
        sys.stderr.write("You must specify Threat Actor with --ta <THREAT ACTOR>\n")
        sys.exit(1)

    # Normalize and lookup country/technique/sectors
    # Print error with suggested match if not found
    ok = True

    try:
        args.target_countries = [
            country_lookup(args, country) for country in args.target_countries
        ]
    except (VocabularyException, FileNotFoundError, worker.UnsupportedScheme) as e:
        sys.stderr.write(str(e))
        ok = False

    try:
        args.techniques = [technique_lookup(tech) for tech in args.techniques]
    except VocabularyException as e:
        sys.stderr.write(str(e))
        ok = False

    try:
        if args.ta_located_in:
            args.ta_located_in = country_lookup(args, args.ta_located_in)
    except VocabularyException as e:
        sys.stderr.write(str(e))
        ok = False

    try:
        args.sectors = [sector_lookup(args, sector) for sector in args.sectors]
    except VocabularyException as e:
        sys.stderr.write(str(e))
        ok = False

    if not ok:
        sys.exit(1)

    if args.target_countries:
        add_ta_target_country(
            actapi, args.output_format, args.ta, args.target_countries
        )

    if args.campaign:
        add_ta_campaign(actapi, args.output_format, args.ta, args.campaign)

    if args.tools:
        add_ta_tools(actapi, args.output_format, args.ta, args.tools)

    if args.techniques:
        add_ta_techniques(actapi, args.output_format, args.ta, args.techniques)

    if args.sectors:
        add_ta_sectors(actapi, args.output_format, args.ta, args.sectors)

    if args.ta_located_in:
        add_ta_located_in(actapi, args.output_format, args.ta, args.ta_located_in)


def main_log_error() -> None:
    "Call main() and log all excetions  as errors"
    try:
        main()
    except Exception:
        error("Unhandled exception: {}".format(traceback.format_exc()))
        raise


if __name__ == "__main__":
    main_log_error()



#!/usr/bin/env python3

"""Worker module polling events from argus incidents"""

import argparse
import os
import socket
import sys
import time
import traceback
from logging import debug, error, info, warning
from typing import Any, Dict, Generator, Optional, Text, cast, List

import caep
import requests
import urllib3

import act.api
from act.workers.libs import argus, mnemonic, worker
from act.api.libs import cli

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

WORKER_NAME = "argus-case"


def parseargs() -> argparse.ArgumentParser:
    """Parse arguments"""
    parser = worker.parseargs("ARGUS Case enrichment")
    parser.add_argument(
        "--argus-baseurl",
        dest="argus_baseurl",
        default="https://api.mnemonic.no/",
        help="Argus API host",
    )
    parser.add_argument(
        "--argus-timeout", dest="timeout", type=int, default=300, help="Timeout"
    )
    parser.add_argument(
        "--content-props",
        default="file.sha256,process.sha256",
        help="Comma separated list of properties that represents a content object",
    )
    parser.add_argument(
        "--hash-props",
        default="file.md5,process.md5,file.sha1,process.sha1,file.sha512,process.sha512",
        help="Comma separated list of properties that represents a hash",
    )
    parser.add_argument(
        "--last-update",
        type=int,
        help="Last updated timestamp (epoc, seconds). Default - use now-1w first time, and start on last retrieved event at next run.",
    )
    parser.add_argument("--argus-apikey", dest="argus_apikey", help="Argus API key")
    parser.add_argument(
        "--organization-from-argus",
        action="store_true",
        help="Apply organization from argus event",
    )

    return parser


def event_case_query(
    argus_baseurl: Text,
    apikey: Text,
    last_update: int,
    timeout: int,
    proxy_string: Optional[Text] = None,
) -> Generator[Dict[str, Any], None, None]:
    """Query the argus for events associated to cases.
    argus_baseurl - the url to the ARGUS api (https://api.mnemonic.no)
    apikey - Argus API key
    timeout - timeout towards API
    proxy_string - proxy string for the request
    """

    # Batch size
    limit = 2000

    # Use start time 1w prior to lastUpdated timestamp
    # Events can have "old" startTimestamp if they are delayed into argus.
    # lastUpdateTimstamp = timestamp when event is added to Argus
    # startTimeTimstamp = original timestamp of the event
    start_time = last_update - 3600 * 24 * 7 * 1000

    try:
        criteria = {
            "lastUpdatedTimestamp": last_update,
            "startTimestamp": start_time,
            "sortBy": ["lastUpdated"],
            "includeFlags": ["NOTIFIED"],
            "limit": limit,
            # Do not include events that are associated to incidents from streaming filter
            "excludeFlags": ["ASSOCIATED_TO_CASE_BY_FILTER"],
        }

        headers = {"Argus-API-Key": apikey}

        # Do query in batches and yield events
        yield from mnemonic.batch_query(
            "POST",
            "{}/events/v1/aggregated/search".format(argus_baseurl.rstrip("/")),
            headers=headers,
            timeout=timeout,
            json_params=criteria,
            proxy_string=proxy_string,
        )

    except worker.ServiceTimeout as err:
        warning("Service timeout: {}".format(err))

    except (
        urllib3.exceptions.ReadTimeoutError,
        requests.exceptions.ReadTimeout,
        socket.timeout,
    ) as err:
        error("Timeout ({0.__class__.__name__})".format(err))


def get_last_update() -> int:
    "Get last update from disk (~/.cache/<worker_name>/last_update)"
    cache_filename: Text = os.path.join(
        caep.get_cache_dir(WORKER_NAME, create=True), "last_update"
    )

    if os.path.isfile(cache_filename):
        # Read last_update from last recorded succsfully recieved event
        with open(cache_filename) as f:
            last_update = int(f.read().strip())
            debug("last update starting at {}".format(last_update))
    else:
        # last_update not specified, set to now-1w
        last_update = int((time.time() - 3600 * 24 * 7) * 1000)
        info("last update not specified, autoconfigured as {}".format(last_update))

    return last_update


def update_last_update(last_update: int) -> None:
    "Write last update from disk (~/.cache/<worker_name>/last_update)"
    cache_filename: Text = os.path.join(
        caep.get_cache_dir(WORKER_NAME, create=True), "last_update"
    )

    # Write last update timestamp to disk
    with open(cache_filename, "w") as f:
        f.write(str(last_update))


# pylint: disable=too-many-arguments
def process(api: act.api.Act, args: argparse.Namespace) -> None:
    """Get events associated to cases since last update"""

    last_update: Optional[int] = args.last_update
    content_props: List[Text] = [prop.strip() for prop in args.content_props.split(",")]
    hash_props: List[Text] = [prop.strip() for prop in args.hash_props.split(",")]

    # Last update is not specified, get last_update from file (cache)
    if not last_update:
        last_update = get_last_update()

    # Get events
    for counter, event in enumerate(
        event_case_query(
            args.argus_baseurl,
            args.argus_apikey,
            last_update,
            timeout=args.timeout,
            proxy_string=args.proxy_string,
        )
    ):

        # Result is sorted by lastUpdateTimestamp, so we update
        # last_update from the event
        last_update = cast(int, event["lastUpdatedTimestamp"])

        if args.organization_from_argus:
            shortName = event.get("customerInfo", {}).get("shortName")

            if shortName:
                api.config.organization = shortName
            else:
                error(f"Unable to get organization from event: {event}")
                continue

        # Create facts from event
        argus.handle_argus_event(
            api, event, content_props, hash_props, args.output_format
        )

        # For every Nth event, update last updated event
        if (counter % 1000) == 0:
            info(
                "Offset: {}, last_update: {}".format(
                    counter, time.asctime(time.localtime(last_update / 1000))
                )
            )
            update_last_update(last_update)

    # Make sure last_update is updated
    update_last_update(last_update)


def main() -> None:
    "main function"

    # Look for default ini file in "/etc/actworkers.ini" and
    # ~/config/actworkers/actworkers.ini
    # (or replace .config with $XDG_CONFIG_DIR if set)

    args = cli.handle_args(parseargs())

    if not args.argus_apikey:
        cli.fatal("You must specify --argus-apikey on command line or in config file")

    actapi = worker.init_act(args)
    process(actapi, args)


def main_log_error() -> None:
    "Main function. Log all exceptions to error"
    try:
        main()
    except Exception:
        error("Unhandled exception: {}".format(traceback.format_exc()))
        raise


if __name__ == "__main__":
    main_log_error()
